/* PC-6001 emulator on terminal */
/* AKIKAWA, Hisashi 2015.10.13  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <wchar.h>
#include <locale.h>
#include <editline/readline.h>
#include "z80.h"

#if defined __CYGWIN__ || defined __MSYS__
# include <w32api/windows.h>
#endif

#ifdef __MINGW32__
# include "win.h"
  extern WINDOW *stdscr;
#else
# define _XOPEN_SOURCE_EXTENDED
# include <ncursesw/ncurses.h>
# include <termios.h>
#endif

#ifdef X11
# include <X11/XKBlib.h>
# include <X11/keysym.h>
Display *dispX11 = NULL;
#endif

#define DEBUG 1
#define CLOCK_4M 3993600
//#define CLOCK_14M 14.318e6
#define CLOCK_14M (4.5e6 / 286 * 910)
#define TICK (910 * 262 / CLOCK_14M)

/* 8255 status */
#define NOBF	0x80	/* PC7 = #output buffer full   = 8049 #INT */
#define INTE1	0x40	/* PC6 = write interrupt enable            */
#define IBF	0x20	/* PC5 = input buffer full     = 8049 T0   */
#define INTE2	0x10	/* PC4 = read interrupt enable             */
#define INTR	0x08	/* PC3 = interrupt request                 */
#define CGSW	0x04	/* cgrom switch                            */
#define BSRQEN	0x02	/* bus request enable                      */
#define PRTSTRB	0x01	/* printer strobe                          */

#define SETBIT(a) {port[0x92] |= (a);}
#define RESBIT(a) {port[0x92] &= ~(a);}

#define NONE -1

byte readmem(word address, int *wait);
void writemem(word address, byte value, int *wait);
byte readio(word address);
void writeio(word address, byte value);
void init(int argc, char *argv[]);
void reset_p6(void);
int upd8049(int reset);
byte upd8049read(void);
void upd8049write(byte value);
void upd8049cmd(int cmd);
void upd8255(void);
void interrupt(int intr);
void keyinput(int *inkeyp, int *gamekeyp);
int keyvector(int inkey, int *keycode);
void display(void);
char *getinput(int y, const char *prompt);
int upperhex(int value, int mask);
FILE *read_open(const char *fname);
int getp6t(int *data);
void chgvram(void);
void wait60hz(void);
extern void monitor(void);
extern void monitor_refresh(void);

FILE *fp = NULL;
FILE *tape_read = NULL;
FILE *tape_write = NULL;
int normal_speed = 1;
int use_extram = 1;
int use_extrom = 0;	/* 16KB ROM */
int use_beluga = 0;	/* 128KB ROM + 64KB RAM */
int ram64 = 0;
int turbo = 1;
int boost = 8;
int winkey = 0;
int x11key = 0;
long long count_cycles = 0;
int break_address = NONE;
int break_read = NONE;
int break_write = NONE;
int break_in = NONE;
int break_out = NONE;
int break_intr = NONE;
int break_memory = 0;
int break_io = 0;
int count_break = 0;
int until_break= 1;
int step = 0;
int tape_counter;
wint_t keybuf[16384];
int ikbf, okbf;

z80 cpu;
byte ram[0x10000];
byte rom[0x4000];
byte extrom[0x80000];
byte extram[0x20000];
byte cgrom[0x1000];
byte port[0x100];

int page;
int relay;
int timer;
byte *vram;
int count_timer;

/* for 8255 */
int nrd;		/* #read                        */
int nwr;		/* #write                       */
int nack;		/* PC6 = #acknoledge = 8049 #RD */
int nstb;		/* PC4 = #strobe     = 8049 #WR */
byte ibuffer, obuffer;

/* for 8049 */
enum {
  GAME,
  GAMEDATA,
  IDLE,
  SENDKEY,
  READ,
  WRITEOPEN,
  WRITEDATA,
};
int mode_8049;
int count_query;
int count_tape;
int baudrate_read;
int baudrate_write;
int kana_mode;
int katakana_mode;

/* for P6T format */
int count_auto = NONE;
int is_p6t, p6t_block, p6t_blank, p6t_iauto, p6t_waittape;
struct {
  byte version;
  byte nblock;
  byte autostart;
  byte mode;
  byte pages;
  word nauto;
  byte *autostring;
  word nextend;
  byte **data;
} p6t;

/* display character */
const wchar_t dispchar[]
= L"　月火水木金土日年円時分秒百千万"
  "π┻┳┫┣╋┃━┏┓┗┛×大中小"
  "　！”＃＄％＆’（）＊＋，‐．／"
  "０１２３４５６７８９：；＜＝＞？"
  "＠ＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯ"
  "ＰＱＲＳＴＵＶＷＸＹＺ［￥］＾＿"
  "　ａｂｃｄｅｆｇｈｉｊｋｌｍｎｏ"
  "ｐｑｒｓｔｕｖｗｘｙｚ｛｜｝〜　"
  "▲▼★◆○●をぁぃぅぇぉゃゅょっ"
  "　あいうえおかきくけこさしすせそ"
  "　。「」、・ヲァィゥェォャュョッ"
  "ーアイウエオカキクケコサシスセソ"
  "タチツテトナニヌネノハヒフヘホマ"
  "ミムメモヤユヨラリルレロワン゛゜"
  "たちつてとなにぬねのはひふへほま"
  "みむめもやゆよらりるれろわん　　";


int main(int argc, char *argv[])
{
  int i;
  double count_display;
  const double period = 910 / CLOCK_14M;

  srand(time(NULL));
  count_display = TICK * rand() / RAND_MAX;

  init(argc, argv);
  reset_p6();

  while (1) {
    /* 8255, 8049, Z80 */
    for (i = count_display / period; i < 262; i++) {
      count_display = count_display - (int)(count_display / period) * period;
      if (i >= 70 && port[0x92] & BSRQEN) {
	cpu.cycles += (37 / (CLOCK_14M / 16) - count_display) * CLOCK_4M;
	count_display += (double)cpu.cycles / CLOCK_4M;
      }

      while (count_display < period) {
	upd8255();
	interrupt(upd8049(0));
	count_cycles += cpu.cycles;
	count_display += (double)cpu.cycles / CLOCK_4M;

	z80_main(&cpu);
	count_cycles += cpu.cycles;
	count_display += (double)cpu.cycles / CLOCK_4M;

	if ((cpu.pc == break_address || break_memory || break_io)
	    && ++count_break >= until_break) {
	  count_break = 0;
	  monitor();
	} else if (step == 1) {
	  monitor();
	}
	break_memory = 0;
	break_io = 0;
      }
      count_display = count_display - (int)(count_display / period) * period;
    }

    if (port[0x92] & BSRQEN) {
      display();
    }

    if (normal_speed && (relay == 0 || turbo == 0)) {
      wait60hz();
    }

  }
}


byte readmem(word address, int *wait)
{
  int bank;
  int dummy;

  wait = wait ? wait: &dummy;
  *wait = cpu.m1 ? 1 : 0;

  if ((port[0x92] & CGSW) == 0 && 0x6000 <= address && address <= 0x6fff) {
    *wait = 1;
    return cgrom[address - 0x6000];
  }

  if (use_beluga) {	  /* Beluga cartridge */
    if (address < 0x8000) {
      *wait = 1;
    }
    if (address < 0x4000
	&& (port[0x06] != 0x66
	    || (port[0xf0] & 0x0f) == 1
	    || (port[0xf0] & 0x0f) == 2
	    || (port[0xf0] & 0x0f) == 5
	    || (port[0xf0] & 0x0f) == 6)) {
      return rom[address];
    } else if (address >= 0xc000) {
      return ram[address];
    } else {
      bank = port[0x30 + (address >> 13)];
      switch (bank & 0xc0) {
      case 0x00:	/* external ROM */
	//      bank = romset + (bank & 0xcf);
	return extrom[((bank << 13) & 0x7e000) + (address & 0x1fff)];
      case 0x40:	/* external RAM */
	return extram[((bank << 13) & 0x1e000) + (address & 0x1fff)];
      case 0x80:	/* SCC */
	return 0xff;
      case 0xc0:	/* not implemented */
	return 0xff;
      default:
	return 0xff;
      }
    }
  } else if (ram64) {		  /* 64KB RAM mode */
    if (((port[0xf0 + ((address >> 15) & 1)] >> ((address >> 12) & 4)) & 0x0f) == 0x0d) {
      return ram[address];
    } else {
      *wait = 1;
      return rom[address];
    }
  } else {			/* normal PC-6001 mode */
    if (address < 0x8000) {
      *wait = 1;
      if (address < 0x4000) {
	return rom[address];
      } else {
	return extrom[address - 0x4000];
      }
    } else {
      if (address < 0xc000) {
	return extram[address];
      } else {
	return ram[address];
      }
    }
  }
}


void writemem(word address, byte value, int *wait)
{
  int dummy;

  wait = wait ? wait : &dummy;
  *wait = 0;

  if (use_beluga) {	  /* Beluga cartridge */
    //    int bank = romset + (port[0x30 + (address >> 13)] & 0xcf);
    int bank = port[0x30 + (address >> 13)];

    if (address < 0x8000) {
      *wait = 1;
      if (port[0x06] == 0x66
	  && ((port[0xf2] >> ((address >> 13) & 6)) & 3)
	  && ((bank & 0xc0) == 0x40)) {
	extram[((bank << 13) & 0x1e000) + (address & 0x1fff)] = value;
      }
    } else if (address < 0xc000) {
      if ((bank & 0xc0) == 0x40) {
	extram[((bank << 13) & 0x1e000) + (address & 0x1fff)] = value;
      }
    } else {
      ram[address] = value;
    }
  } else if (ram64) {		/* 64KB RAM mode */
    if (((port[0xf2] >> ((address >> 13) & 6)) & 3) == 1) {
      ram[address] = value;
    } else {
      *wait = 1;
    }
  } else {
    if (address < 0x8000) {
      *wait = 1;
    } else if (address < 0xc000) {
      if (use_extram) {
	extram[address] = value;
      }
    } else {
      ram[address] = value;
    }
  }
}


byte readio(word address)
{
  address &= 0xff;

  switch (address) {
//case 0xc0:return 0;
//case 0x81:return 0;
  case 0x90:
  case 0x94:
  case 0x98:
  case 0x9c:
    nrd = 0;
    upd8255();
    nrd = 1;
    upd8255();
    return ibuffer;

  case 0x92:
  case 0x93:
  case 0x96:
  case 0x97:
  case 0x9a:
  case 0x9b:
  case 0x9e:
  case 0x9f:
    return port[0x92];

  case 0xb0:
  case 0xb1:
  case 0xb2:
  case 0xb3:
  case 0xb4:
  case 0xb5:
  case 0xb6:
  case 0xb7:
  case 0xb8:
  case 0xb9:
  case 0xba:
  case 0xbb:
  case 0xbc:
  case 0xbd:
  case 0xbe:
  case 0xbf:
    writeio(0xb0, 0xff);
    return 0xff;

  default:
    //    return port[address];
    return 0xff;

  }
}


void writeio(word address, byte value)
{
  FILE *serial;

  address &= 0xff;

  switch (address) {

  case 0x80:
    if ((serial = fopen("serial.txt", "ab")) != NULL) {
      fputc(value, serial);
      fclose(serial);
    }
    break;

  /* Beluga cartridge */
  case 0x06:
    port[address] = value;
    break;

  case 0x30:
  case 0x31:
  case 0x32:
  case 0x33:
  case 0x34:
  case 0x35:
  case 0x36:
  case 0x37:
  case 0x38:
  case 0x39:
  case 0x3a:
  case 0x3b:
  case 0x3c:
  case 0x3d:
  case 0x3e:
  case 0x3f:
    //    beluga[address & 0x07] = value;
    port[address & 0xf7] = value;
    port[address | 0x08] = value;
    if ((value & 0xc0) == 0) {
      //      romset = value & 0x30;
      for (address = 0x30; address < 0x38; address++) {
	if ((port[address] & 0xc0) == 0) {
	  port[address] = (value & 0x30) + (port[address] & 0x0f);
	  port[address + 0x08] = port[address];
	}
      }
    }
    chgvram();
    break;
    
  case 0x70:
  case 0x71:
  case 0x72:
  case 0x73:
  case 0x74:
  case 0x75:
  case 0x76:
  case 0x77:
  case 0x78:
  case 0x79:
  case 0x7a:
  case 0x7b:
  case 0x7c:
  case 0x7d:
  case 0x7e:
  case 0x7f:
    port[address] = value;
    port[0x32] = (port[0x32] & 0x30) + (value & 0x0f);
    break;

  /* 8255 port A = 8049 data */
  case 0x90:
  case 0x94:
  case 0x98:
  case 0x9c:
    nwr = 0;
    upd8255();
    nwr = 1;
    upd8255();
    obuffer = value;
    break;

  /* 8255 port B = printer data */
  case 0x91:
  case 0x95:
  case 0x99:
  case 0x9d:
    port[0x91] = value;
    break;

  /* 8255 control register */
  case 0x93:
  case 0x97:
  case 0x9b:
  case 0x9f:
    if (value & 0x80) {			/* 8255 mode set */
      ;					/* not implemented */
    } else {				/* 8255 bit set/reset */
      if (value & 0x01) {
	SETBIT(1 << ((value & 0x0e) >> 1));
      } else {
	RESBIT(1 << ((value & 0x0e) >> 1));
      }
      upd8255();
    }
    break;

  case 0xb0:
  case 0xb1:
  case 0xb2:
  case 0xb3:
  case 0xb4:
  case 0xb5:
  case 0xb6:
  case 0xb7:
  case 0xb8:
  case 0xb9:
  case 0xba:
  case 0xbb:
  case 0xbc:
  case 0xbd:
  case 0xbe:
  case 0xbf:
    /* timer */
    if (!timer && !(value & 0x01)) {	/* off -> on */
      count_timer = 8192 / 2;
    }
    timer = !(value & 0x01);

    /* change VRAM address */
    page = (value & 0x06) >> 1;		/* !!! 0-3 !!! */
    chgvram();

    /* tape relay */
    if (relay && !(value & 0x08)) {	/* on -> off */
      p6t_waittape = 0;
    }
    relay = (value & 0x08) >> 3;

    break;

  case 0xf0:
  case 0xf1:
  case 0xf2:
    if (ram64 | use_beluga) {
      port[address] = value;
    }
    break;
  }
}


byte z80_read(word address)
{
  byte value;
  int wait;

  if (address == break_read) {
    break_memory = 1;
  }

  value = readmem(address, &wait);
  if (wait) {
    cpu.cycles++;
  }

  return value;
}


void z80_write(word address, byte value)
{
  int wait;

  if (address == break_write) {
    break_memory = 1;
  }

  writemem(address, value, &wait);
  if (wait) {
    cpu.cycles++;
  }
}


byte z80_in(word address)
{
  if ((address & 0xff) == break_in) {
    break_io = 1;
  }
  if ((address & 0xf0) == 0xa0) {
    cpu.cycles++;
  }
  return readio(address);
}


void z80_out(word address, byte value)
{
  if ((address & 0xff) == break_out) {
    break_io = 1;
  }
  if ((address & 0xf0) == 0xa0) {
    cpu.cycles++;
  }
  writeio(address, value);
}


/*
 * initialize emulator
 */
void init(int argc, char *argv[])
{
  int i;
  int rawmode;
  FILE *rom_file;
  FILE *extrom_file;
  extern int rl_completion_append_character;

  if (DEBUG) {
    fp = fopen("log", "w");
  }

#ifdef __MINGW32__
  /* Windows key input */
  rawmode = 1;
  winkey = 1;
#else
  rawmode = 0;
#endif
  
  /* analyze option */
  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-raw") == 0) {
      rawmode = 1;
    } else if (strcmp(argv[i], "-noraw") == 0) {
      rawmode = 0;
    } else if (strcmp(argv[i], "-16") == 0) {
      use_extram = 0;
    } else if (strcmp(argv[i], "-64") == 0) {
      ram64 = 1;
    } else if (strcmp(argv[i], "-nowait") == 0) {
      normal_speed = 0;
    } else if (strcmp(argv[i], "-noturbo") == 0) {
      turbo = 0;
    } else if (strcmp(argv[i], "-boost") == 0) {
      if (++i >= argc) {
	break;
      }
      sscanf(argv[i], "%d", &boost);
      if (boost < 1) {
	boost = 1;
      }
    } else if (strcmp(argv[i], "-win") == 0) {
#ifdef WIN32
      winkey = 1;
#else
      printf("-win option is only for Windows\n");
      exit(EXIT_FAILURE);
#endif
    } else if (strcmp(argv[i], "-x11") == 0) {
#ifdef X11
      x11key = 1;
#else
      printf("-x11 option is only for X11\n");
      exit(EXIT_FAILURE);
#endif
    } else if (strcmp(argv[i], "-tape") == 0) {
      if (++i >= argc) {
	break;
      }
      if ((tape_read = read_open(argv[i])) == NULL) {
	printf("cannot open %s\n", argv[i]);
	exit(EXIT_FAILURE);
      }
      if (is_p6t && p6t.autostart) {
	if (p6t.mode == 1 || p6t.mode == 3) {
	  use_extram = 0;
	} else {
	  use_extram = 1;
	}
      }
    } else if (strcmp(argv[i], "-rom") == 0) {
      if (++i >= argc) {
	break;
      }

      memset(extrom, 0xff, sizeof(extrom));
      if ((extrom_file = fopen(argv[i], "rb")) == NULL) {
	printf("cannot open %s\n", argv[i]);
	exit(EXIT_FAILURE);
      }
      if (fread(extrom, 1, sizeof(extrom), extrom_file) <= 0x4000) {
	use_extrom = 1;
      } else {
	use_beluga = 1;
      }
      fclose(extrom_file);
    } else {
      printf("option:\n"
	     " -16              16KB RAM\n"
	     " -64              64kB RAM (mkII mode2)\n"
	     " -nowait          start with full speed\n"
	     " -noturbo         no turbo tape\n"
	     " -boost <n>       tape speed=1200*n baud (default=8)\n"
	     " -x11             use X11 API for game key input\n"
	     " -win             use Windows API for game key input\n"
	     " -raw             raw mode TTY\n"
	     " -noraw           not raw mode TTY\n"
	     " -tape <filename> set tape file\n"
	     " -rom <filename>  set external rom file\n"
	     );
      exit(EXIT_FAILURE);
    }
  }

#ifdef X11
  /* X11 key input */
  if (x11key) {
    if ((dispX11 = XOpenDisplay(NULL)) == NULL) {
      printf("cannot open display\n");
      exit(EXIT_FAILURE);
    }
  }
#endif

  /* set RAM & ROM */
  memset(ram, 0x00, sizeof(ram));
  if (use_extram == 0 && ram64 == 0) {
    memset(&ram[0x8000], 0xff, 0x4000);
  }
  memset(rom, 0xff, sizeof(rom));
  memset(cgrom, 0, sizeof(cgrom));
  memset(extram, 0xff, sizeof(extram));

  rom_file = fopen("BASICROM.60", "rb");
  if (rom_file == NULL) {
    printf("cannot open BASICROM.60\n");
    exit(EXIT_FAILURE);
  }
  fread(rom, 1, sizeof(rom), rom_file);
  fclose(rom_file);

  rom_file = fopen("CGROM60.60", "rb");
  if (rom_file) {
    fread(cgrom, 1, sizeof(cgrom), rom_file);
    fclose(rom_file);
  }

  /* readline */
  using_history();
  rl_completion_append_character = '\0';

  /* ncurses */
  setlocale(LC_ALL, "");
  initscr();
  nonl();
  noecho();
  keypad(stdscr, TRUE);
  nodelay(stdscr, TRUE);
  curs_set(0);
  scrollok(stdscr, TRUE);
  set_escdelay(20);
  if (rawmode) {
    raw();
  }
}


/*
 * reset hardware
 */
void reset_p6()
{
  int i;

  tape_counter = 0;
  if (tape_read) {
    fseek(tape_read, 0, SEEK_SET);
  }
  if (is_p6t) {
    p6t_block = 0;
    p6t_blank = 1;
    p6t_iauto = -2;
    p6t_waittape = 0;
    if (p6t.autostart) {
      if (count_auto != NONE || p6t_iauto > p6t.nauto) {
	count_auto = NONE;
      } else {
	count_auto = 0;
      }
    }
  }

  memset(port, 0xff, sizeof(port));
  writeio(0xb0, 0);

  count_timer = rand() / (RAND_MAX / 8192 + 1);

  if (ram64 || use_beluga) {
    port[0xf0] = 0x71;	/* 0000-3fff:internal ROM, 4000-7fff:external ROM */
    port[0xf1] = 0xdd;	/* 8000-ffff:internal RAM */
    port[0xf2] = 0x55;  /* 0000-ffff:internal RAM for writing */
    if (use_beluga) {
      port[0x06] = 0xff;
      for (i = 0; i < 8; i++) {
	port[0x30 + i] = 0;
      }
    }
  }

  z80_reset(&cpu);

  upd8049(1);

  /* 8255 */
  port[0x92] = NOBF | INTE1 | INTE2 | INTR;
  nrd = 1;
  nwr = 1;
  nack = 1;
  nstb = 1;
  ibuffer = 0xff;
  obuffer = 0xff;
}


/*
 * uPD8049 (sub CPU)
 *  return 1=interrupt, 0=no interrupt
 */
int upd8049(int reset)
{
  int period;
  static int cmd;
  static int count_key;
  static int count_next;
  static int port90_next;
  static int inkey;
  static int gamekey;
  int data;

  /* reset */
  if (reset) {
    mode_8049 = IDLE;
    cmd = NONE;
    kana_mode = 0;
    katakana_mode = 0;
    baudrate_read = 1200;
    baudrate_write = 1200;
    count_tape = NONE;
    ikbf = 0;
    okbf = 0;
    inkey = ERR;

    /* reset keyinput */
    gamekey = 0xff;
    keyinput(&inkey, &gamekey);

    return 0;
  }

  count_key += cpu.cycles;
  period =  TICK * CLOCK_4M;
  if (count_key > period) {
    count_key %= period;
    keyinput(&inkey, &gamekey);
  }

  switch (mode_8049) {

  /* reply for game key query (in 8049 interrupt routine?) */
  case GAME:
    count_query += cpu.cycles;
    if (count_query > 500 && (port[0x92] & IBF) == 0) {
      port90_next = gamekey;
      count_next = 0;
      upd8049write(0x16);
      mode_8049 = GAMEDATA;
      return 1;
    }
    break;

  /* send game key data */
  case GAMEDATA:
    count_next += cpu.cycles;
    if (port[0x92] & IBF) {
      return 1;
    }
    if (count_next > 100) {
      upd8049write(port90_next);
      mode_8049 = IDLE;
    }
    break;

  /* key input */
  case IDLE:
    if (cmd != NONE) {
      upd8049cmd(cmd);
      cmd = NONE;
    } else if ((port[0x92] & NOBF) == 0) {
      cmd = upd8049read();
    } else if (inkey != ERR && (port[0x92] & IBF) == 0) {
      upd8049write(keyvector(inkey, &port90_next));
      mode_8049 = SENDKEY;
      count_next = 0;
      inkey = ERR;
      return 1;
    }
    break;

  /* send key code and CMT read data*/
  case SENDKEY:
    count_next += cpu.cycles;
    if (cmd == NONE && (port[0x92] & NOBF) == 0) {
      cmd = upd8049read();
    } else if (port[0x92] & IBF) {
      return 1;
    } else if (count_next > 100) {
      upd8049write(port90_next);
      if (count_tape == NONE) {
        mode_8049 = IDLE;
      } else {
	mode_8049 = READ;
      }
      inkey = ERR;
    }
    break;

  /* CMT read */
  case READ:
    if (inkey == KEY_BREAK || inkey == KEY_END) {
      /* CMT read stop */
      upd8049cmd(0x1a);
      upd8049write(0x10);
      return 1;
    } else if ((port[0x92] & NOBF) == 0) {
      cmd = upd8049read();
      if (cmd != 0x06) {
	upd8049cmd(cmd);
	cmd = NONE;
      }
      inkey = ERR;
      break;
    }
    inkey = ERR;

    if (getp6t(&data)) {
      mode_8049 = SENDKEY;
      port90_next = data;
      count_next = 0;
      upd8049write(0x08);
      if (port90_next != EOF) {
	tape_counter++;
	mvprintw(16, 74, "%6d", tape_counter);
      }
      return 1;
    }
    break;

  /* CMT write */
  case WRITEOPEN:
  case WRITEDATA:
    count_tape += cpu.cycles;
    period = CLOCK_4M * 12 / baudrate_write / boost;
    if (count_tape > period) {
      count_tape %= period;
      if (mode_8049 == WRITEDATA && (port[0x92] & NOBF) == 0) {
	mode_8049 = WRITEOPEN;
	if (tape_write) {
	  fputc(upd8049read(), tape_write);
	  fflush(tape_write);
	}
      }
    }

    if (inkey == KEY_BREAK || inkey == KEY_END) {
      /* CMT write stop */
      upd8049cmd(0x3a);
      upd8049write(0x0e);
      inkey = ERR;
      return 1;
    }
    inkey = ERR;

    if (mode_8049 == WRITEOPEN && (port[0x92] & NOBF) == 0) {
      cmd = upd8049read();
      if (cmd != 0x06) {
	upd8049cmd(cmd);
	cmd = NONE;
      }
      break;
    }

    break;
  }

  return 0;
}


/*
 * uPD8049 reads data from uPD8255
 */
byte upd8049read()
{
  nack = 0;
  upd8255();
  nack = 1;
  upd8255();
  return obuffer;
}


/*
 * uPD8049 writes data to uPD8255
 */
void upd8049write(byte value)
{
  nstb = 0;
  upd8255();
  nstb = 1;
  upd8255();
  ibuffer = value;
}


/*
 * uPD8049 executes command
 */
void upd8049cmd(int cmd)
{
  switch (cmd) {

  case 0x06:		/* game key query */
    mode_8049 = GAME;
    count_query = 0;
    break;

  case 0x19:		/* open CMT for reading */
    mode_8049 = READ;
    count_tape = 0;
    break;

  case 0x1d:		/* set baud rate for reading (600 baud) */
    baudrate_read = 600;
    break;

  case 0x1e:		/* set baud rate for reading (1200 baud) */
    baudrate_read = 1200;
    break;

  case 0x38:		/* write to CMT */
    if (mode_8049 == WRITEOPEN) {
      mode_8049 = WRITEDATA;
    }
    break;

  case 0x39:		/* open CMT for writing */
    tape_write = fopen("savedata.p6", "ab");
    mode_8049 = WRITEOPEN;
    break;

  case 0x3d:		/* set baud rate for writing (600 baud) */
    baudrate_write = 600;
    break;

  case 0x3e:		/* set baud rate for writing (1200 baud) */
    baudrate_write = 1200;
    break;

  case 0x1a:		/* close CMT for reading */
  case 0x3a:		/* close CMT for writing */
    mode_8049 = IDLE;
    count_tape = NONE;
    break;

  case 0x0c:		/* RS-232C ? */
  case 0x34:		/* ???? */
    break;
  }
}


/*
 * uPD8255 (PPI)
 */
void upd8255()
{
  FILE *printer;
  static byte port92_old = 0;
  static int nrd_old = 1, nwr_old = 1;
  static int nack_old = 1, nstb_old = 1;
  static int print_data;

  if (nack_old && nack == 0) {		/* H -> L */
    SETBIT(NOBF);
  }
  if (nstb_old && nstb == 0) {		/* H -> L */
    SETBIT(IBF);
  }
  if ((port92_old & PRTSTRB) && (port[0x92] & PRTSTRB) == 0) {	/* H -> L */
    if ((printer = fopen("printer.txt", "ab")) != NULL) {
      fputc((~print_data) & 0xff, printer);
      fclose(printer);
    }
  }
  if ((port92_old & PRTSTRB) == 0 && (port[0x92] & PRTSTRB)) {	/* L -> H */
    print_data = port[0x91];
  }
  if (nrd_old == 0 && nrd) {					/* L -> H */
    RESBIT(IBF);
  }
  if (nwr_old == 0 && nwr) {					/* L -> H */
    RESBIT(NOBF);
  }

  /* interrupt(PC3): refered to uPD8255A data sheet */
  if ((nrd && nstb && (port[0x92] & IBF) && (port[0x92] & INTE2))
      || (nwr && nack && (port[0x92]  & NOBF) && (port[0x92] & INTE1))) {
    SETBIT(INTR);
  } else {
    RESBIT(INTR);
  }

  port92_old = port[0x92];
  nrd_old = nrd;
  nwr_old = nwr;
  nack_old = nack;
  nstb_old = nstb;
}


/*
 * interrupt
 */
void interrupt(int intr)
{
  int period = 8192;
  int vector = NONE;

  /* 2ms timer */
  count_timer += cpu.cycles;
  cpu.cycles = 0;

  if (cpu.iff1 && ! cpu.ei) {
    if (intr) {
      vector = z80_in(0x90);
    } else if (timer && count_timer >= period) {
      count_timer %= period;
      vector = 0x06;
    }

    if (vector != NONE) {
      z80_int(&cpu, vector);
      count_timer += cpu.cycles;
      if ((cpu.pc == break_address || vector == break_intr)
	  && ++count_break >= until_break) {
	count_break = 0;
	monitor();
      } else if (step == 1) {
	monitor();
      }
    }
  }

}


/*
 * key input
 */
void keyinput(int *inkeyp, int *gamekeyp)
{
  static int i;
  static int period;
  wint_t c;

  if (*gamekeyp == 0xff) {	/* reset */
    period = 80;
    *gamekeyp = 0;
    return;
  }

  if (count_auto != NONE && mode_8049 == IDLE && !p6t_waittape) {
    count_auto++;
    if (count_auto >= period) {
      count_auto = 0;
      if (p6t_iauto == -2) {
	*inkeyp = '0' + p6t.pages;
	p6t_iauto++;
      } else if (p6t_iauto == -1) {
	*inkeyp = 0x0d;
	p6t_iauto++;
      } else {
	*inkeyp = p6t.autostring[p6t_iauto++];
	if (*inkeyp == 0x0a) {
	  *inkeyp = 0x0d;
	  p6t_waittape = 1;
	}
	if (p6t_iauto >= p6t.nauto) {
	  count_auto = NONE;
	}
      }
      if (*inkeyp == 0x0d) {
	period = 40;
      } else {
	period = 1;
      }
      return;
    }
  }  

  i++;
  if (i % 12 == 0) {
    *gamekeyp = 0;
  }
  while (get_wch(&c) != ERR) {
    if (c == KEY_F(24)) {				/* exit */
      endwin();
      exit(EXIT_SUCCESS);
    } else if (c == KEY_F(6)) {				/* monitor mode */
      monitor();
    } else if (c == KEY_F(10)) {			/* no wait mode */
      normal_speed = !normal_speed;
    } else if (c == KEY_F(11) || c== KEY_F(23)) {	/* reset */
      reset_p6();
      *inkeyp = ERR;
    } else if (c == KEY_F(9)) {				/* set tape file */
      char *fname, *fname_expand;
      int y = getmaxy(stdscr) - 1;
      fname = getinput(y, "tape file: ");
      if (strlen(fname) > 0) {
	fname_expand = tilde_expand(fname);
	if (tape_read = read_open(fname_expand)) {
	  mvprintw(y, 0, "open %s", fname_expand);
	  tape_counter = 0;
	} else {
	  mvprintw(y, 0, "cannot open %s", fname_expand);
	}
	clrtoeol();
	free(fname_expand);
      }
      free(fname);
    } else if (c == KEY_F(21)) {			/* set extrom file */
      char *fname, *fname_expand;
      FILE *extrom_file;
      int y = getmaxy(stdscr) - 1;
      fname = getinput(y, "external rom file: ");
      if (strlen(fname) > 0) {
	fname_expand = tilde_expand(fname);
	if (extrom_file = fopen(fname_expand, "rb")) {
	  mvprintw(y, 0, "open %s", fname_expand);
	  if (fread(extrom, 1, sizeof(extrom), extrom_file) <= 0x4000) {
	    use_extrom = 1;
	  } else {
	    use_beluga = 1;
	  }
	  fclose(extrom_file);
	  reset_p6();
	} else {
	  mvprintw(y, 0, "cannot open %s", fname_expand);
	}
	clrtoeol();
	free(fname_expand);
      }
      free(fname);
    } else if ((ikbf + 1) % (sizeof(keybuf) / sizeof(keybuf[0])) != okbf) {
      keybuf[ikbf = (ikbf + 1) % (sizeof(keybuf) / sizeof(keybuf[0]))] = c;

      switch (c) {
      case ' ':			/* space */
	*gamekeyp |= 0x80;
	break;
      case KEY_LEFT:		/* left */
	*gamekeyp |= 0x20;
	break;
      case KEY_RIGHT:		/* right */
	*gamekeyp |= 0x10;
	break;
      case KEY_DOWN:		/* down */
	*gamekeyp |= 0x08;
	break;
      case KEY_UP:		/* up */
	*gamekeyp |= 0x04;
	break;
      case KEY_BREAK:		/* stop */
      case KEY_END:		/* stop */
	*gamekeyp |= 0x02;
	break;
      case 'z':
      case 'Z':
	if (! x11key && ! winkey) {
	  *gamekeyp |= 0x01;
	}
	break;
      }
    }
  }
  if (*inkeyp == ERR) {
    *inkeyp = (ikbf == okbf ? ERR : keybuf[okbf = (okbf + 1) % (sizeof(keybuf) / sizeof(keybuf[0]))]);
  }

#ifdef WIN32
  if (winkey) {
    *gamekeyp
      = ((GetKeyState(' ')      < 0) ? 0x80 : 0)
      | ((GetKeyState(VK_LEFT)  < 0) ? 0x20 : 0)
      | ((GetKeyState(VK_RIGHT) < 0) ? 0x10 : 0)
      | ((GetKeyState(VK_DOWN)  < 0) ? 0x08 : 0)
      | ((GetKeyState(VK_UP)    < 0) ? 0x04 : 0)
      | ((GetKeyState(VK_END)   < 0) ? 0x02 : 0)
      | ((GetKeyState(VK_SHIFT) < 0) ? 0x01 : 0);
  }
#endif

#ifdef X11
  if (x11key) {
    char keys[32];
    KeyCode code;

    *gamekeyp = 0;
    XQueryKeymap(dispX11, keys);

    code = XKeysymToKeycode(dispX11, XK_space);
    if (keys[code / 8] & (1 << (code % 8))) {
      *gamekeyp |= 0x80;
    }
    code = XKeysymToKeycode(dispX11, XK_Left);
    if (keys[code / 8] & (1 << (code % 8))) {
      *gamekeyp |= 0x20;
    }
    code = XKeysymToKeycode(dispX11, XK_Right);
    if (keys[code / 8] & (1 << (code % 8))) {
      *gamekeyp |= 0x10;
    }
    code = XKeysymToKeycode(dispX11, XK_Down);
    if (keys[code / 8] & (1 << (code % 8))) {
      *gamekeyp |= 0x08;
    }
    code = XKeysymToKeycode(dispX11, XK_Up);
    if (keys[code / 8] & (1 << (code % 8))) {
      *gamekeyp |= 0x04;
    }
    code = XKeysymToKeycode(dispX11, XK_End);
    if (keys[code / 8] & (1 << (code % 8))) {
      *gamekeyp |= 0x02;
    }
    code = XKeysymToKeycode(dispX11, XK_Break);
    if (keys[code / 8] & (1 << (code % 8))) {
      *gamekeyp |= 0x02;
    }
    code = XKeysymToKeycode(dispX11, XK_Shift_L);
    if (keys[code / 8] & (1 << (code % 8))) {
      *gamekeyp |= 0x01;
    }
    code = XKeysymToKeycode(dispX11, XK_Shift_R);
    if (keys[code / 8] & (1 << (code % 8))) {
      *gamekeyp |= 0x01;
    }
  }
#endif
}


/*
 * return interrupt vector and convert key code
 */
int keyvector(int inkey, int *keycode)
{
  int i;
  const int key_special[][3] = {
    {0x02, 0x00, KEY_F(7)},		/* KANA key */
    {0x02, 0x00, KEY_F(8)},		/* shift + PAGE key */
    {0x02, 0x00, KEY_SPREVIOUS},	/* shift + PAGE key */
    {0x02, 0x09, KEY_BTAB},		/* shift + tab key */
    {0x02, 0x0b, KEY_SHOME},		/* shift + HOME-CLR key */
    {0x02, 0x0c, KEY_HOME},		/* HOME-CLR key */
    {0x02, 0x12, KEY_IC},		/* INS key */
    {0x02, 0x1c, KEY_RIGHT},		/* cursor right */
    {0x02, 0x1c, KEY_SRIGHT},		/* shift + cursor right */
    {0x02, 0x1d, KEY_LEFT},		/* cursor left */
    {0x02, 0x1d, KEY_SLEFT},		/* shift + cursor left */
    {0x02, 0x1e, KEY_UP},		/* cursor up */
    {0x02, 0x1e, KEY_SR},		/* shift + cursor up */
    {0x02, 0x1f, KEY_DOWN},		/* cursor down */
    {0x02, 0x1f, KEY_SF},		/* shift + cursor down */
    {0x02, 0x7f, KEY_BACKSPACE},	/* DEL key */
    {0x02, 0x7f, KEY_DC},		/* DEL key*/
    {0x02, 0x7f, KEY_SDC},		/* shift + DEL key*/
    {0x02, 0xfe, KEY_PPAGE},		/* PAGE key */
    {0x14, 0xf0, KEY_F(1)},		/* F1 */
    {0x14, 0xf1, KEY_F(2)},		/* F2 */
    {0x14, 0xf2, KEY_F(3)},		/* F3 */
    {0x14, 0xf3, KEY_F(4)},		/* F4 */
    {0x14, 0xf4, KEY_F(5)},		/* F5 */
    {0x14, 0xf5, KEY_F(13)},		/* shift+F1 */
    {0x14, 0xf6, KEY_F(14)},		/* shift+F2 */
    {0x14, 0xf7, KEY_F(15)},		/* shift+F3 */
    {0x14, 0xf8, KEY_F(16)},		/* shift+F4 */
    {0x14, 0xf9, KEY_F(17)},		/* shift+F5 */
    {0x14, 0xfa, KEY_BREAK},		/* STOP key */
    {0x14, 0xfa, KEY_END},		/* STOP key */
  };

  const byte key_convert[][2] = {
    {0x00, 0x00}, {0x01, 0x00}, {0x02, 0x00}, {0x03, 0x00},
    {0x04, 0x00}, {0x05, 0x00}, {0x06, 0x00}, {0x07, 0x00},
    {0x08, 0x00}, {0x09, 0x00}, {0x0a, 0x00}, {0x0b, 0x00},
    {0x0c, 0x00}, {0x0d, 0x00}, {0x0e, 0x00}, {0x0f, 0x00},
    {0x10, 0x00}, {0x11, 0x00}, {0x12, 0x00}, {0x13, 0x00},
    {0x14, 0x00}, {0x15, 0x00}, {0x16, 0x00}, {0x17, 0x00},
    {0x18, 0x00}, {0x19, 0x00}, {0x1a, 0x00}, {0x1b, 0x00},
    {0x1c, 0x00}, {0x1d, 0x00}, {0x1e, 0x00}, {0x1f, 0x00},

    /*               !             "             # ぁ    */
    {0x20, 0x20}, {0x00, 0x00}, {0x00, 0x00}, {0x87, 0x00},
    /* $ ぅ          % ぇ          & ぉ          ' ゃ    */
    {0x89, 0x00}, {0x8a, 0x00}, {0x8b, 0x00}, {0x8c, 0x00},
    /* ( ゅ          ) ょ          *             +       */
    {0x8d, 0x00}, {0x8e, 0x00}, {0x00, 0x00}, {0x00, 0x00},
    /* , ね 小       - ほ ━       . る 大       / め ▲ */
    {0xe8, 0x1f}, {0xee, 0x17}, {0xf9, 0x1d}, {0xf2, 0x80},
    /* 0 わ 万       1 ぬ 日       2 ふ 月       3 あ 火 */
    {0xfc, 0x0f}, {0xe7, 0x07}, {0xec, 0x01}, {0x91, 0x02},
    /* 4 う 水       5 え 木       6 お 金       7 や 土 */
    {0x93, 0x03}, {0x94, 0x04}, {0x95, 0x05}, {0xf4, 0x06},
    /* 8 ゆ 百       9 よ 千       : け ▼       ; れ ★ */
    {0xf5, 0x0d}, {0xf6, 0x0e}, {0x99, 0x81}, {0xfa, 0x82},
    /* < 、          =             > 。          ? ・    */
    {0xa4, 0x00}, {0x00, 0x00}, {0xa1, 0x00}, {0xa5, 0x00},
    /* @ ゛          A             B             C       */
    {0xde, 0x00}, {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00},
    /* D             E    ぃ       F             G       */
    {0x00, 0x00}, {0x88, 0x00}, {0x00, 0x00}, {0x00, 0x00},
    /* H             I             J             K       */
    {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00},
    /* L             M             N             O       */
    {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00},
    /* P             Q             R             S       */
    {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00},
    /* T             U             V             W       */
    {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00}, {0x00, 0x00},
    /* X             Y             Z っ          [ ゜ ○ */
    {0x00, 0x00}, {0x00, 0x00}, {0x8f, 0x00}, {0xdf, 0x84},
    /* \ ー 円       ] む ●       ^ へ          _ ろ ◆ */
    {0xb0, 0x09}, {0xf1, 0x85}, {0xed, 0x00}, {0xfb, 0x83},
    /* `             a ち          b こ ┛       c そ ┗ */
    {0x00, 0x00}, {0xe1, 0x00}, {0x9a, 0x1b}, {0x9f, 0x1a},
    /* d し ┣       e い ┏       f は ╋       g き ┫ */
    {0x9c, 0x14}, {0x92, 0x18}, {0xea, 0x15}, {0x97, 0x13},
    /* h く 時       i に ┃       j ま          k の    */
    {0x98, 0x0a}, {0xe6, 0x16}, {0xef, 0x00}, {0xe9, 0x00},
    /* l り 中       m も 分       n み          o ら    */
    {0xf8, 0x1e}, {0xf3, 0x0b}, {0xf0, 0x00}, {0xf7, 0x00},
    /* p せ π       q た          r す ┳       s と 秒 */
    {0x9e, 0x10}, {0xe0, 0x00}, {0x9d, 0x12}, {0xe4, 0x0c},
    /* t か ┓       u な          v ひ ┻       w て    */
    {0x96, 0x19}, {0xe5, 0x00}, {0xeb, 0x11}, {0xe3, 0x00},
    /* x さ ×       y ん 年       z つ          { 「    */
    {0x9b, 0x1c}, {0xfd, 0x08}, {0xe2, 0x00}, {0xa2, 0x00},
    /* |             } 」          ~                     */
    {0x00, 0x00}, {0xa3, 0x00}, {0x00, 0x00}, {0x7f, 0x7f}
  };

  if (inkey == 0x1b) {
    inkey = (ikbf == okbf ? ERR : keybuf[okbf = (okbf + 1) % (sizeof(keybuf) / sizeof(keybuf[0]))]);
    if (inkey == ERR) {
      *keycode = 0x1b;
      return 0x02;
    }

    if (inkey <= 0x7f) {
      *keycode = key_convert[inkey][1];
      if (*keycode < 0x20) {
	return 0x14;
      } else {
	return 0x02;
      }
    } else {
      *keycode = 0;
      return 0x14;
    }
  }

  if (inkey < 0x80) {
    if (kana_mode) {
      *keycode = key_convert[inkey][0];
      if (katakana_mode
	  && ((0x86 <= *keycode && *keycode <= 0x9f)
	      || (0xe0 <= *keycode && *keycode <= 0xfd))) {
	*keycode ^= 0x20;
      }
    } else {
      *keycode = inkey;
    }
    return 0x02;
  }

  for (i = 0; i < sizeof(key_special) / sizeof(key_special[0]); i++) {
    if (inkey == key_special[i][2]) {
      if (inkey == KEY_F(7)) {
	kana_mode = !kana_mode;
      } else if (inkey == KEY_F(8) || inkey == KEY_SPREVIOUS) {
	katakana_mode = !katakana_mode;
      }
      *keycode = key_special[i][1];
      return key_special[i][0];
    }
  }

  for (i = 0x20; i < 0x100; i++) {
    if(inkey == dispchar[i]) {
      *keycode = i;
      return 0x02;
    }
  }
  for (i = 0; i < 0x20; i++) {
    if(inkey == dispchar[i]) {
      *keycode = i;
      return 0x14;
    }
  }

  *keycode = 0;
  return 0x14;
}


/*
 * display
 */
void display()
{
  int i, j;
  int x, y;
  cchar_t wch;
  int graphic_mode;
  int color_used[4];
  int color1, color2;
  byte pattern[4][12];
  int weight[2];
  int ul, ur, dl, dr;
  int bitmap[128][32];
  char render[65] = {};
  byte table[] =
    " ._.____...___jj..______..______,,__,,,L,,____Ly,,___Lyy,,__Lyyy"
    "...jjjjj.._jjjjj.._jjjju.__jjxuo,,_jjuoy,>vjuyyy,jvuoyyy,cuoyyyy"
    "^.jjjjjj.jjjjjjj//_jjjjj/_jjjunn,,jjjzaa,jjjzyyy,)znayyg/zzjaygg"
    ".^_jjjj]./jjjj]1//jjj]11/>jju11q,jv]11aa,c]n1yyq,]z11yqgYzJwaqgg"
    "^^_jjj]1/jjjj]11/>jj]JJJ}jvu11dd,cJ212y2,zJeyy2ZYzJdydgO7JJddgOG"
    "^^jjj]11/<jj]111/jJ]JJJ3rcJJ1qdd,zJJay22YzJwyggg7JJddgOW7J2d4gG@"
    "^<jjj111/jj]1111/)JJ1J33,cJJdqq4,zJwa2gO7JJygggW7J2dggW@722dOG@@"
    "^jjj111yjj]j1113/JJJ1333YzJdaq4O,zJa2gOG7w22ggW87y2dOW@@72ZOG@@@"
    "''''''<j..___<jx..__{lxu.___lxuo,i_LxuLL,ivxuLyy,lvuLyyy,cLLyyyy"
    "^^^^^<jx.<<j<jxu//v>vxuu/ivxxuos,>vxuyyy}vvuyyyyrcuoyyyyrcoeyyyy"
    "^^^^jjj]///jjj]1//>j)un1/>vxunna/zvuzaaarzznayyg/zzeayggYzCwaggg"
    "^^^]1]]1/<<j1]11/>v]]111}vv]11qqrz]z1yyqY]zeyyqgYzJwyqggfzVyqggW"
    "^^<j1]11/<]]]111/)J]1JJ3rcJd1dddYzJ2ay22YzJwy2gZ7JJddgOG7J2d4OG@"
    "^<J]]111/<J]1111/JJJJJ337JJJdyd4YzJ2a2gO7JJydggW7JJd4gG@7J2dOG@@"
    "^)J]111y/]]]1113/JJJ1333YJJdaq4OzJJa2ggG7Jy2ggW07y2dgG@@72ZOG@@@"
    "^)]]11y3/]J11139YJJJ33357JJddqOG7z22ggG07y2mgWg@72ZOG@@@2ZZW@@@#"
    "''''i<i'!!!i<<iu!ii>i)un:iiicooo:iLLLLLw}iLLLewy:cLLewyg:cLewygg"
    "^^^^<<cu!<<<<<uo/>>))uoo:>vcuoow}ccooewarceoewygrcoewyggczewaggg"
    "^^^<j)nn^<<j)nn1/>))nnna})cnnqaarzzzaaagzzzeaaggYzewagggzzeagggg"
    "^^<]1]11^<)]]111>))]11qqrcnn1qqqYzzeaaqgYzewaqggzzVaqggWfzadggWg"
    "^<)]]11d<)]]111dr)Jn1333YcodddddYz22a222zz22dggG722d4gG@722d4G@@"
    "^)J]1133r)]]1133))JJ3333YzJdaqdOzz222g2G7w22ggW0722dgg@@72ZOG@@@"
    "^)]n11a3)]J11339YJJ3333377VddqqG7z22ggG07222ggg072ZOG@@@2ZZW@@@#"
    "^]J11a34YzJ3aq497JJ3343%77dd4qG87222ag0072Z2Gg@#2ZZW@@@#PZZ@@@#@"
    "'''''l'L!!!{llLL!i{llLLL:llLLLLL:lLLLLLallLLLLyp:LLLLypmYLLLapmm"
    "'^^<((uo!<<j(uos:{((uoCX}lcuossyrcLLswyprILpwyppYLCwypppYkpppppp"
    "^^^j^^CC^<)j]nCX}))nnCXarcCnCXaqYzLCaaggYzCmammmYCmmmmmmTmKmmmmm"
    "^^^]]11V}(]]11VVY(Vn1VV3YYVUVqUUYzVXaUUDYVVpmmWWfVVUUWW@7VmUWWW@"
    "^^]]11V3r(]]1V33YYJ34334Y7VX444O77V24ZOG7ZZpZZG@7ZZZOG@@7ZZOG@@@"
    "^(]313337YJ33339Y7J3333377V344OG77Z24OG07ZZmZW007ZZOG@@@7ZZW@@@#"
    "^YJ33339Y7J3339977V3335G77VU4OG87ZZUOG007pZZG@@@7ZZW@@@@ZZZ@@@#@"
    "Y7CX339977V3399G77V345%87VVUOG807ZZOG0007ZZWG@@#ZZZ@@@@@ZZZ@@#@@"
    "''''''''''<llILh:{llILhh:lLILhhKlLLLLLyblILLwypmILLwypm&LkLppg&W"
    "''<''Ihh'<((uhhh}((uhhhKrIIohhKbIILkwybbIkkwypp&Ikpppp&WkkKpp&W@"
    "^^^junCX}))]1CXKr)CnC%KSYICCKKq6YzCwaggDTkKmmmmWTKKmm&W@TKmm&%@@"
    "^^]]11VSf)]]11SSfYVSSSSSYfVXKq6OfkVKUDDWfVKWWWWWfKmWWW@@fmZWWW@@"
    "'']]1V3Sf]J113S9YfJS3SS5ffVOGOOGfkK2GOGG7ZZOOG@@7ZZOG@@@PZZW@@@@"
    "'IJ99999Yf999999f7V339AG77KOGOG87KZ0OG007ZZOG000PZZW@@@@PAW@@@#@"
    "YfJ99999ff999999f7K3S5G87KKOOG807bZOG000PZZW000#PZW@@@@@PW@@@#@@"
    "ffV9999Gf79999G97K3S5%807K9OG800PZDW000#PZW@@@##PW@@@@@@W@@@#@@@"
    "''''''Lh'{llLLhh:lLhLhhKllLLhhKb:LLLLyb&ILL&yp&&TkL&&&&&TkKb&&&&"
    "'<(xuhhh}<(uhhhKr(hhhhK%TIohhKb6Tkkbab&&Tkp&&&&&TkK&&&&&TKb&&&&&"
    "^^)TTT%KT)CnT%K%TTCTT%%%TTKKK%6%TTKKbg&%TTKm&&%&TKK&&%%@TKm&N%@#"
    "^)]S1SSST]]S1SSSYTVSSSS5TTVKN6NNTFKbNDNNFKm6&NNBFbE&NN@$FDDWN@$#"
    "^YJ55555YTJ55555TTV55555TFKdN5N8FK255NBBPZZ5N#B#PZZ@@@##P5E@@###"
    "YT999999T9999999T7K555N57KR9NN80PPZDN000PZAW#0##PAE00###PRR0####"
    "TT99995NT99995N9FK955%5BPP95N8BBPPDBBBB#P&EB####PERB####PRBB####"
    "TFK995NNF9995%NBPP95N%BBPPA88BB#PDEBB###PERB####PRBB####PBB#####"
    "'''''Lhh!lLLL%%%llLLh%%%:ILhhK%%ILLLab&&LkLpp&&%TkKp&&%%FKb&&%%%"
    "'(vuh%%%r(Ihh%%%rIhhh%%%IIhhK%%%Tkbbb&&%TkKp&&%&FKp&&%%@Fb&&&%%$"
    "^)Cn1%%%TTCT%%%%TTCF%%%%TTKKq%%%TFKbg%%%FKmm&%%%Fmm&%%%$FmE&%%$%"
    "^Y]11%S%YFVSS%%%FFVSS%%%FFKUN%NNFEEEDEEEFEEDN%$$FEEEN@$#PEEE@$#$"
    "YFJ55%55FFV55%5%FFK555%MFPROO%RBPPEOGRB$PZEW#B$#PEE@@@##PER@####"
    "PP99999%PF9999%MFPP95%MMPPRANR00PPEE000#PREM0###PER0####PRR0####"
    "FFK99%%MFP995%MMPP95N%%%PPRMR00#PEE00###PER0####PRR$####PR$#####"
    "FP995%%%PP95N%%%PP5MM%%%PRRMBB##PERB####PRR$####P@$#####P$######"
    "''''LhhK!lLLh%%%:ILhh%%%ILLhK%%%LkLbb&&%TkKp&&%$FKp&&%$$Fpm&&%$$"
    "'(Ioh%%%rIChh%%%YThha%%%TkKKb%%%TkKb&&%$Fpp&&%$$Fpp&&%$$F&&&$$$$"
    "^^CCa%%%TTCXa%%%TTKK%%%%TFK66%%%FKK6&%%$Fmm&%%$$FmE$$$$$P&E$$$$$"
    "YfVSS%S%ffVSS%%%fFKSS%%%FKK6N%$$FEEDE$$$F$$$$$$$P$E$$$$$$E$$$$$$"
    "ffV5455%fFK555%MFP555%%$PPROGR$$PZEE$$$$P$E$$$$$PE$@$$$$P$$$#$$$"
    "7F9999%MF9999%M0PP95N%0$PPR880$#PEE00$#$PER$$#$$PR$$#$$$P$$$#$$$"
    "FP995%M$PP99N%$$PP55M%$$PRR80$$$PER0$#$$P$$$#$$$$$$$#$$$$$$$#$$$"
    "PP95N%$$P999M%$$P5RM$$$$PRR0$$$$ER$$#$$$$$$$#$$$$$$$$$$$$$$$$$$$";

  graphic_mode = 0;
  for (i = 0; i < 0x200; i++) {
    if (vram[i] & 0x80) {
      graphic_mode = 1;
      break;
    }
  }

  if (graphic_mode) {
    /* graphic mode
      (x,y) PC-6001 text coordinates
      (2x,y) terminal coordinates
      (4x,2y) AA framebuffer coordinates
      2*6 P6 mode4 dots in framebuffer pixel
    */

    for (y = 0; y < 16; y++) {
      for (x = 0; x < 32; x++) {
	for (i = 0; i < 4; i++) {
	  weight[0] = 0;
	  weight[1] = 0;
	  for (j = 0; j < 12; j++) {
	    weight[j / 6] +=
	      ((vram[0x200 + x + (y * 12 + j) * 32] >> (6 - i * 2)) & 0x03)
	      * ((vram[x + ((y * 12 + j) % 16) * 32] & 0x02) ? 1 : 2);
	  }
	  bitmap[x * 4 + i][y * 2 + 0] =
	    (weight[0] == 0) ? 0 : (weight[0] + 6) / 6;
	  bitmap[x * 4 + i][y * 2 + 1] =
	    (weight[1] == 0) ? 0 : (weight[1] + 6) / 6;
	}
	for (i = 0; i < 2; i++) {
	  ul = bitmap[x * 4 + i * 2    ][y * 2    ];
	  ur = bitmap[x * 4 + i * 2 + 1][y * 2    ];
	  dl = bitmap[x * 4 + i * 2    ][y * 2 + 1];
	  dr = bitmap[x * 4 + i * 2 + 1][y * 2 + 1];
	  render[x * 2 + i] = table[(ul << 9) + (ur << 6) + (dl << 3) + dr];
	}
      }
      mvprintw(y, 0, "%s", render);

      /* screen mode 4 text character */
      for (x = 0; x < 32; x++) {
	for (i = 0; i < 0x100; i++) {
	  for (j = 0; j < 12; j++) {
	    if (vram[0x200 + y * 384 + j * 32 + x] != cgrom[i * 16 + j]) {
	      break;
	    }
	  }
	  if (j == 12) {
	    if (i) {
	      setcchar(&wch, &dispchar[i], A_NORMAL, 0, NULL);
	      mvadd_wch(y, x * 2, &wch);
	    }
	    break;
	  }
	  for (j = 0; j < 12; j++) {
	    if (vram[0x200 + y * 384 + j * 32 + x] != (~cgrom[i * 16 + j] & 0xff)) {
	      break;
	    }
	  }
	  if (j == 12) {
	    if (i) {
	      setcchar(&wch, &dispchar[i], A_REVERSE, 0, NULL);
	      mvadd_wch(y, x * 2, &wch);
	    }
	    break;
	  }
	}
      }

      /* screen mode 3 text character */
      for (x = 0; x < 16; x++) {
	memset(color_used, 0, sizeof(color_used));
	memset(pattern, 0, sizeof(pattern));

	for (i = 0; i < 12; i++) {
	  for (j = 0; j < 4; j++) {
	    color1 = (vram[0x200 + y * 384 + i * 32 + x * 2]
		      >> (6 - j * 2)) & 0x03;
	    color2 = (vram[0x200 + y * 384 + i * 32 + x * 2 + 1]
		      >> (6 - j * 2)) & 0x03;
	    pattern[color1][i] += 1 << (7 - j);
	    pattern[color2][i] += 1 << (3 - j);
	    color_used[color1] = 1;
	    color_used[color2] = 1;
	  }
	}
	if (color_used[0] + color_used[1] + color_used[2] + color_used[3] != 2) {
	  continue;
	}
	color1 = color_used[0] ? 0 : color_used[1] ? 1 : 2;
	color2 = color_used[3] ? 3 : color_used[2] ? 2 : 1;
	for (i = 0; i < 0x100; i++) {
	  for (j = 0; j < 12; j++) {
	    if (pattern[color2][j] != cgrom[i * 16 + j]) {
	      break;
	    }
	  }
	  if (j < 12) {
	    for (j = 0; j < 12; j++) {
	      if (pattern[color1][j] != cgrom[i * 16 + j]) {
		break;
	      }
	    }
	  }
	  if (j == 12) {
	    mvaddch(y, x * 4, ' ');
	    setcchar(&wch, &dispchar[i], A_NORMAL, 0, NULL);
	    mvadd_wch(y, x * 4 + 1, &wch);
	    mvaddch(y, x * 4 + 3, ' ');
	    break;
	  }
	}
      }
    }

  } else {
    /* semi-graphic mode
      (x,y) PC-6001 text coordinates
      (2x,y) terminal coordinates
      (4x,2y) AA framebuffer coordinates
      0.5*1.5 P6 mode2 dots in framebuffer pixel
    */
    for (y = 0; y < 16; y++) {
      for (x = 0; x < 32; x++) {
	for (i = 0; i < 2; i++) {
	  weight[0] = ((vram[0x200 + x + y * 32] >> (4 - i)) & 2)
	    + ((vram[0x200 + x + y * 32] >> (3 - i)) & 1);
	  weight[1] = ((vram[0x200 + x + y * 32] >> (3 - i)) & 1)
	    + ((vram[0x200 + x + y * 32] << i) & 2);
	  ul = weight[0] * (((vram[x + y * 32] & 0x02) << 1) + ((vram[0x200 + x + y * 32] & 0xc0) >> 6)) / 3;
	  ur = ul;
	  dl = weight[1] * (((vram[x + y * 32] & 0x02) << 1) + ((vram[0x200 + x + y * 32] & 0xc0) >> 6)) / 3;
	  dr = dl;
	  render[x * 2 + i] = table[(ul << 9) + (ur << 6) + (dl << 3) + dr];
	}
      }
      mvprintw(y, 0, "%s", render);
    }

    /* screen mode 1 and 2 text character */
    for (i = 0; i < 0x200; i++) {
      if ((vram[i] & 0x40) == 0) {
	x = i & 0x1f;
	y = (i & 0x1e0) >> 5;
	if (vram[i] & 0x01) {
	  setcchar(&wch, &dispchar[vram[0x200 + i]], A_REVERSE, 0, NULL);
	} else {
	  setcchar(&wch, &dispchar[vram[0x200 + i]], A_NORMAL, 0, NULL);
	}
	mvadd_wch(y, x * 2, &wch);
      }
    }
  }
}


/*
 * print prompt and get input by readline/editline
 */
char *getinput(int y, const char *prompt)
{
  int i, j;
  int maxy;
  char *input;
  static char *input_prev = NULL;

#ifndef __MINGW32__
  struct termios term;
  tcgetattr(0, &term);
  term.c_lflag |= ECHO;
  tcsetattr(0, TCSANOW, &term);
#endif

  maxy = getmaxy(stdscr);
  setscrreg(19, maxy - 1);
  if (y >= maxy) {
    y = maxy - 1;
  }
  move(y, 0);
  clrtoeol();
  curs_set(1);
  refresh();

  do {
    input = readline(prompt);
    redrawwin(stdscr);
    refresh();
  } while (input == NULL);

#ifndef __MINGW32__
  term.c_lflag &= ~ECHO;
  tcsetattr(0, TCSANOW, &term);
#endif

  curs_set(0);

  maxy = getmaxy(stdscr);
  setscrreg(19, maxy - 1);

  mvprintw(y, 0, "%s%s", prompt, input);
  i = strspn(input, " ");
  for (j = strlen(input) - 1; j > i; j--) {
    if (input[j] != ' ') {
      break;
    }
  }
  memmove(input, input + i, j - i + 1);
  input[j -i + 1] = '\0';
  if (strlen(input) > 0 && (input_prev == NULL || strcmp(input, input_prev))) {
    add_history(input);
    if (input_prev) {
      free(input_prev);
    }
    input_prev = malloc(strlen(input) + 1);
    strcpy(input_prev, input);
  }

  return input;
}


/*
 * get the upper hex digits
 */

int upperhex(int value, int mask)
{
  if (value < 0) {
    value = -value;
    while (value > mask) {
      value >>= 4;
    }
    value = -value;
  }
  while (value > mask) {
    value >>= 4;
  }

  return value;
}


/*
 * tape read file open
 */
FILE *read_open(const char *fname)
{
  int i;
  int c;
  FILE *new;

  if (fname == NULL || (new = fopen(fname, "rb")) == NULL) {
    return NULL;
  }
  
  if (tape_read) {
    fclose(tape_read);
    tape_read = NULL;
  }

  if (is_p6t) {
    free(p6t.autostring);
    if (p6t.nblock) {
      for (i = 0; i < p6t.nblock; i++) {
	free(p6t.data[i]);
      }
      free(p6t.data);
    }
  }

  is_p6t = 0;
  if (strcasecmp(&fname[strlen(fname) - 4], ".p6t")) {
    return new;
  }

  fseek(new, -4, SEEK_END);
  fread(&i, 4, 1, new);
  fseek(new, i, SEEK_SET);
  c = fgetc(new);
  if (c == 'P' && fgetc(new) == '6') {
    p6t.version = fgetc(new);
    p6t.nblock = fgetc(new);
    p6t.autostart = fgetc(new);
    p6t.mode = fgetc(new);
    p6t.pages = fgetc(new);
    fread(&p6t.nauto, 2, 1, new);
    p6t.autostring = malloc(p6t.nauto + 1);
    for (i = 0; i < p6t.nauto; i++) {
      p6t.autostring[i] = fgetc(new);
    }
    p6t.autostring[i] = '\0';
    fread(&p6t.nextend, 2, 1, new);
    for (i = 0; i < p6t.nextend; i++) {
      fgetc(new);
    }
    if (p6t.nblock) {
      p6t.data = malloc(sizeof(p6t.data[0]) * p6t.nblock);
      for (i = 0; i < p6t.nblock; i++) {
	p6t.data[i] = malloc(33);
	fread(p6t.data[i], 1, 33, new);
      }
    }
    fseek(new, 4, SEEK_CUR);
    if (i == p6t.nblock && fgetc(new) == EOF) {
      is_p6t = 1;
      p6t_block = 0;
      p6t_blank = 1;
    }
  }

  fseek(new, 0, SEEK_SET);
  tape_counter = 0;
  return new;
}


/*
 * get data from .p6t file
 */
int getp6t(int *data)
{
  static int idata;
  int period;

  if (!tape_read) {
    return 0;
  }

  count_tape += cpu.cycles;

  if (is_p6t && p6t_blank) {
    if (p6t_block >= p6t.nblock) {
      *data = EOF;
      count_tape = 0;
      return 1;
    }
    period = CLOCK_4M / 1000
      * (p6t.data[p6t_block][21] + (p6t.data[p6t_block][22] << 8)
	 + p6t.data[p6t_block][23] + (p6t.data[p6t_block][24] << 8));
  } else {
    /* start bit(1) + data(8) + stop bit(3) = 12 */
    period = CLOCK_4M * 12 / baudrate_read;
    if (boost > 1) {
      if (ram[0xfa19]) {
	count_tape = 0;
      } else {
	period /= boost;
      }
    }
  }

  if (count_tape <= period) {
    return 0;
  }

  if (is_p6t) {
    if (p6t_blank) {
      p6t_blank = 0;
      count_tape = 0;
      idata = 0;
    }
    idata++;
    if (idata > p6t.data[p6t_block][29] + (p6t.data[p6t_block][30] << 8)
	+ (p6t.data[p6t_block][31] << 16) + (p6t.data[p6t_block][32] << 24)) {
      count_tape = 0;
      p6t_blank = 1;
      p6t_block++;
      if (p6t_block < p6t.nblock) {
	fseek(tape_read,
	      p6t.data[p6t_block][25]
	      + (p6t.data[p6t_block][26] << 8)
	      + (p6t.data[p6t_block][27] << 16)
	      + (p6t.data[p6t_block][28] << 24),
	      SEEK_SET);
      }
      return 0;
    }
  }
    
  *data = fgetc(tape_read);
  count_tape %= period;
  return 1;
}


/*
 * synchronize for 60Hz
 */
void wait60hz()
{
  struct timespec now, wait = {0, TICK / 10 * 1e9};
  static struct timespec old = {0, 0};
  long long tdiff;

  while (1) {
    clock_gettime(CLOCK_MONOTONIC, &now);
    tdiff = now.tv_nsec - old.tv_nsec + (now.tv_sec - old.tv_sec) * 1e9;
    if (tdiff > TICK * 1e9) {
      break;
    }
    nanosleep(&wait, NULL);
  }
  old.tv_sec = now.tv_sec;
  old.tv_nsec = now.tv_nsec - tdiff % (long)(TICK * 1e9);
  if (old.tv_nsec < 0) {
    old.tv_sec--;
    old.tv_nsec += 1e9;
  }
}


/*
 * change VRAM
 */
void chgvram()
{
  int bank;
  word vramtable[] = {0xc000, 0xe000, 0x8000, 0xa000};

  /* !!! page=0-3 !!! */

  if (ram64 || page < 2) {
    vram = &ram[vramtable[page]];
  } else if (!use_beluga) {
    vram = &extram[vramtable[page]];
  } else {
    /* Beluga cartridge */
    bank = port[0x32 + page];	/* 0x34 or 0x35 */
    switch (bank & 0xc0) {
    case 0x00:	/* external ROM */
      vram = &extrom[(bank << 13) & 0x7e000];
      break;
    case 0x40:	/* external RAM */
      vram = &extram[(bank << 13) & 0x1e000];
      break;
    case 0x80:	/* SCC */
      break;
    case 0xc0:	/* not implemented */
      break;
    }
  }
}
